import tensorflow as tf
import os

class RegModelLogistic:
    """ A general implementation of a TensorFlow logistic regression model. """

    verbose = False

    w = None
    b = None

    dataModelColCount = 0
    dataModelAnsColCount = 0
    totalTrainingSteps = 1000
    trainStepsMultiplier = 1
    checkpoint = False
    randomSeed = False
    learning_rate = 0.0000001
    positiveResult = 0.50

    loadTensorData = None
    logPrint = 100
    evalType = ''
    checkpointSave = 100
    callbackMethod = None
    exeCfg = None

    def __init__(self, lVerbose=False, lLoadTensorData=None, lPositiveResult=0.50, lRandomSeed=False, lTrainStepsMultiplier=1.0, lLogPrint=100, lLearningRate=0.0000001, lEvalType=''):
        self.verbose = lVerbose
        self.learning_rate = lLearningRate
        self.loadTensorData = lLoadTensorData
        self.positiveResult = lPositiveResult
        self.randomSeed = lRandomSeed
        self.trainStepsMultiplier = lTrainStepsMultiplier
        self.dataModelColCount = self.loadTensorData.dataModelColCount
        self.dataModelAnsColCount = self.loadTensorData.dataModelAnsColCount
        self.logPrint = lLogPrint
        self.evalType = lEvalType
    # edef

    def setCheckPointSave(self, cps):
        self.checkpointSave = cps;
    # edef

    def setCheckPointOn(self, cpo):
        self.checkpoint = cpo;
    # edef

    def setCallbackMethod(self, mthd):
        self.callbackMethod = mthd;
    # edef

    def setExeCfg(self, cfg):
        self.exeCfg = cfg
    # edef

    # START TF NN CODE

    def combine_inputs(self, x):
        return tf.matmul(x, self.w) + self.b
    # edef

    def inference(self, x):
        # Compute inference model over data x and return the result.
        return tf.nn.softmax(self.combine_inputs(x))
    # edef

    def loss(self, x, y):
        # Compute loss over training data x and expected outputs y.
        return tf.reduce_mean(-tf.reduce_sum(y * tf.log(self.inference(x)), reduction_indices=1))
    # edef

    def inputs(self):
        # Read/generate input training data x and expected outputs y.
        return self.loadTensorData.train, self.loadTensorData.trainAnswers
    # edef

    def train(self, totalLoss):
        return tf.train.GradientDescentOptimizer(self.learning_rate).minimize(totalLoss)
    # edef

    def evaluate(self, sess, test_x, test_y):
        Y_predicted = tf.equal(tf.argmax(self.inference(test_x), 1), tf.argmax(test_y, 1))
        mse = tf.reduce_mean(tf.cast(Y_predicted, tf.float32))
        print('Accuracy: %.4f' % sess.run(mse))
    # edef

    def predict(self, sess, test_x):
        Y_predicted = self.inference(test_x)

        if self.verbose:
            print('PredictionArgMax: %s' % sess.run(tf.argmax(Y_predicted, 1)))
            print('Prediction: %s' % sess.run(Y_predicted))
        # eif

        return sess.run(Y_predicted)
    # edef

    def predictColumn(self, sess, test_x):
        return sess.run(tf.argmax(self.predict(sess, test_x), 1))
    # eif

    def interpretPredictColumn(self, sess, test_x):
        if float(self.predictColumn(sess, test_x)) == 0.0:
            return True
        # eif

        return False
    # eif

    def startTraining(self):

        self.totalTrainingSteps = int(self.loadTensorData.trainCount * self.trainStepsMultiplier)
        print('Found training steps: %i' % self.totalTrainingSteps)

        with tf.Session() as sess:
            print('Found tensor dimension: %ix%i' % (self.dataModelColCount, self.dataModelAnsColCount))
            if self.randomSeed == True:
                self.w = tf.Variable(tf.random_normal([self.dataModelColCount, self.dataModelAnsColCount], stddev=0.5), name='weights')
                self.b = tf.Variable(tf.random_normal([self.dataModelAnsColCount], stddev=0.5), name='bias')
            else:
                self.w = tf.Variable(tf.zeros([self.dataModelColCount, self.dataModelAnsColCount]), name='weights')
                self.b = tf.Variable(tf.zeros([self.dataModelAnsColCount]), name='bias')
            # eif

            print("Weight: %s" % self.w.get_shape())
            print("Bias: %s" % self.b.get_shape())

            # Model setup
            tf.global_variables_initializer().run()

            # Create a saver
            if self.checkpoint == True:
                saver = tf.train.Saver()
            # eif

            x, y = self.inputs()
            total_loss = self.loss(x, y)
            train_op = self.train(total_loss)

            coord = tf.train.Coordinator()
            threads = tf.train.start_queue_runners(sess, coord)
            training_steps = self.totalTrainingSteps

            initial_step = 0

            if self.checkpoint == True:
                # Verify we don't have a checkpoint saved already
                ckpt = tf.train.get_checkpoint_state(os.path.dirname(__file__) + "/checkpoints/")

                if ckpt and ckpt.model_checkpoint_path:
                    # Restores from checkpoint
                    saver.restore(sess, ckpt.model_checkpoint_path)
                    initial_step = int(ckpt.model_checkpoint_path.rsplit('-', 1)[1])
                # eif
            # eif

            # Training loop
            if self.checkpoint == True:
                for step in range(initial_step, training_steps):
                    sess.run([train_op])

                    if step % self.logPrint == 0:
                        print ("Loss: ", sess.run([total_loss]))
                    # eif

                    if step % self.checkpointSave == 0:
                        saver.save(sess, './checkpoints/eod-model', global_step=step)
                    # eif

                # efl
            else:
                for step in range(training_steps):
                    sess.run([train_op])

                    if step % self.logPrint == 0:
                        print ("Loss: ", sess.run([total_loss]))
                    # eif

                # efl
            # eif

            self.evaluate(sess, self.loadTensorData.validate, self.loadTensorData.validateAnswers)

            print('Custom Evaluation: %s' % (self.evalType))
            if self.evalType == 'blue_ice':
                print("Test1: [Bad] %s" % self.interpretPredictColumn(sess, [[127.0, 1.0, 0.0, 1.0, 0.0, 400.0, 0.0]]))
                print("Test2: [Good] %s" % self.interpretPredictColumn(sess, [[127.0, 2.0, 0.0, 1.0, 1.0, 200.0, 0.0]]))
                print("Test3: [Bad] %s" % self.interpretPredictColumn(sess, [[127.0, 3.0, 0.0, 1.0, 0.0, 400.0, 0.0]]))
                print("Test4: [Good] %s" % self.interpretPredictColumn(sess, [[127.0, 4.0, 0.0, 1.0, 1.0, 200.0, 0.0]]))
            # eif

            if self.callbackMethod:
                self.callbackMethod(self, sess, self.evalType, self.exeCfg)
            # eif

        # ewith
    # edef

# eclass
